/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.5)

   class      : CalendarEntry
   date       : 1.10.97
   description: a class that describes a complete calendar entry
   modified   : 11.02.98 - data encryption added
                15.02.98 - method makePeriodicInstances added
                16.02.98 - method reconstructParent added

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.io.*;
import java.util.*;

/* ******************************************************************
   top level class declaration
*********************************************************************/

public class CalendarEntry implements Streamable, Cloneable {

  // instance variable declarations
  final static long INVALID_DATE      = 0;
  final static int  NOT_PERIODICAL    = 0;
  final static int  EVERY_DAY         = 1;
  final static int  EVERY_WEEK        = 2;
  final static int  EVERY_MONTH       = 3;
  final static int  EVERY_MONTH_EXACT = 4;       // if day exists
  final static int  EVERY_MONTH_MATCH = 5;       // with matching day of week
  final static int  NOTES_TAG         = 6;       // contains notes to periodical
                                                 // dates 
  final static int  DELETED_INSTANCE  = 7;       // a periodic date where a
                                                 // single instance has been
                                                 // deleted
  Date startTime;                      // set to INVALID_DATE, if entry is invalid
  Date endTime;
  Date alertTime;                      // set to INVALID_DATE, if no alert
  String description;
  String creator;                      // creator account
  Vector persons;                      // list of persons visiting the meeting
  String location;
  String notes;
  private int periodical;              // NOT_PERIODICAL, EVERY_DAY ... MONTH
  int distance;                        // number of days, weeks or month
                                       // between two periodical entries
  Date expirTime;                      // set to INVALID_DATE, if periodical
                                       // entry never expires
  boolean isPrivateFlag;
  private CalendarEntry oldVersion;    // a pointer to a old version of this
                                       // entry, if one exists
  private boolean lastInSeq;           // indicates that this object is the
                                       // last in a sequence - for networking
                                       // purpose
  private Date perStartTime = null;    // to reconstruct the start time of a
                                       // periodic date
  private Date perEndTime   = null;    // to reconstruct the end time of a
                                       // periodic date
  private Date perAlertTime = null;    // to reconstruct the alert time of a
                                       // periodic date

  // constructors
  public CalendarEntry() {
    init();
  } // end constructor CalendarEntry

  public CalendarEntry(String creator) {
    init();
    this.creator = creator;
  } // end constructor CalendarEntry

  public CalendarEntry (DataInput in) throws IOException {

    startTime     = new Date(INVALID_DATE);
    endTime       = new Date(INVALID_DATE);
    alertTime     = new Date(INVALID_DATE);
    persons       = new Vector();
    expirTime     = new Date(INVALID_DATE);
    inputFrom(in);
  } // end constructor CalendarEntry

  // methods
  private void init() {
    startTime     = new Date(INVALID_DATE);
    endTime       = new Date(INVALID_DATE);
    alertTime     = new Date(INVALID_DATE);
    description   = "";
    creator       = "";
    persons       = new Vector();
    location      = "";
    notes         = "";
    periodical    = NOT_PERIODICAL;
    distance      = 1;
    expirTime     = new Date(INVALID_DATE);
    isPrivateFlag = false;
    oldVersion    = null;
    lastInSeq     = false;
  } // end init

  public String toString() {
    return description;
  } // end toString

  public void outputTo(DataOutput out) throws IOException {
    Cypher cypher = new DataCypher();
    int n         = persons.size();
    int i;

    out.writeLong(startTime.getTime());
    out.writeLong(endTime.getTime());
    out.writeLong(alertTime.getTime());
    out.writeUTF(cypher.cypher(description));
    out.writeUTF(creator);
    for (i=0; i<n; i++) out.writeUTF((String)persons.elementAt(i));
    out.writeUTF("");
    out.writeUTF(cypher.cypher(location));
    out.writeUTF(cypher.cypher(notes));
    out.writeInt(periodical);
    out.writeInt(distance);
    out.writeLong(expirTime.getTime());
    out.writeBoolean(isPrivateFlag);
    out.writeBoolean(oldVersion != null);
    if (oldVersion != null) oldVersion.outputTo(out);
    out.writeBoolean(lastInSeq);
  } // end outputTo

  public void inputFrom(DataInput in) throws IOException {
    Cypher cypher = new DataCypher();
    String str;

    startTime.setTime(in.readLong());
    endTime.setTime(in.readLong());
    alertTime.setTime(in.readLong());
    description = cypher.decypher(in.readUTF());
    creator     = in.readUTF();
    str         = in.readUTF();
    persons     = new Vector();
    while (!str.equals("")) {
      persons.addElement(str);
      str = in.readUTF();
    } // end while;
    location   = cypher.decypher(in.readUTF());
    notes      = cypher.decypher(in.readUTF());
    periodical = in.readInt();
    distance   = in.readInt();
    expirTime.setTime(in.readLong());
    isPrivateFlag = in.readBoolean();
    if (in.readBoolean()) oldVersion = new CalendarEntry(in);
      else oldVersion = null;
    lastInSeq = in.readBoolean();
  } // end inputFrom

  public void setLastInSeq(boolean lastInSeq) {
  // sets if object is the last in a sequence or not

    this.lastInSeq = lastInSeq;
  } // end setLastInSeq

  public boolean isLastInSeq() {
  // indicates if object is the last in a sequence or not

    return lastInSeq;
  } // end isLastInSeq

  public void deleteOn(DataOutput out) throws IOException {

    out.writeLong((new Date(INVALID_DATE)).getTime());
  } // end deleteOn

  public boolean isDeleted() {

    return (startTime.getTime() == INVALID_DATE);
  } // end isDeleted

  public void setDeleted() {
    startTime.setTime(INVALID_DATE);
  } // end setDeleted

  public Date getStart() {

    return startTime;
  } // end getStart

  public void setStart(Date date) {
    startTime = new Date(date.getTime());
  } // setStart

  public Date getEnd() {

    return endTime;
  } // end getEnd

  public void setEnd(Date date) {
    endTime = new Date(date.getTime());
  } // setEnd

  public String getDescription() {
    return description;
  } // end getDescription

  public void setDescription(String str) {
    description = str;
  } // end setDescription

  public String getCreator() {

    return creator;
  } // end getCreator

  public void setCreator(String creator) {

    this.creator = creator;
  } // end setCreator

  public String[] getPersons() {
    int i;
    int n = persons.size();
    String[] aStr = new String[n];

    for (i=0; i<n; i++) aStr[i] = (String)persons.elementAt(i);
    return aStr;
  } // end getPersons

  public boolean hasPerson(String person) {
    return (persons.indexOf(person) >= 0);
  } // end hasPerson

  public void setPersons(String[] persons) {
    int i;
    int n = persons.length;

    this.persons.removeAllElements();
    for (i=0; i<n; i++) this.persons.addElement(persons[i]);
  } // end getPersons

  public String getLocation() {
    return location;
  } // end getLocation

  public void setLocation(String str) {
    location = str;
  } // end setLocation

  public String getNotes() {

    return notes;
  } // end getNotes

  public void setNotes(String str) {

    notes = str;
  } // end setNotes

  public int getPeriodic() {

    return periodical;
  } // end getPeriodic

  public void setPeriodic(int i) {

    periodical = i;
  } // end getPeriodic

  public boolean isPeriodic() {

    return (periodical != NOT_PERIODICAL) &&
           (periodical != DELETED_INSTANCE) &&
           (periodical != NOTES_TAG);
  } // end isPeriodic

  public boolean isDeletedInstance() {

    return (periodical == DELETED_INSTANCE);
  } // end isDeletedInstance

  public boolean deletedInstanceMatches(CalendarEntry entry) {
    return
      startTime.equals(entry.startTime) &&
      endTime.equals(entry.endTime) &&
      alertTime.equals(entry.alertTime) &&
      description.equals(entry.description) &&
      location.equals(entry.location) &&
      (isPrivateFlag == entry.isPrivateFlag);
  } // end deletedInstanceMatches

  private CalendarEntry makeInstance(Date start, Date end) {
    CalendarEntry entry = (CalendarEntry)this.clone();
    entry.perStartTime = entry.startTime;
    entry.perEndTime   = entry.endTime;
    entry.perAlertTime = entry.alertTime;
    entry.setStart(start);
    entry.setEnd(end);
    if (hasAlertTime())
      entry.alertTime = new Date(entry.startTime.getTime() -
                                 (entry.perStartTime.getTime() -
                                  entry.perAlertTime.getTime()));
    return entry;
  } // end makeInstance

  public Enumeration makePeriodicInstances(Date from, Date to) {
    Vector vector = new Vector();
    Date expirTime;

    if (this.expirTime.getTime() == INVALID_DATE)
      expirTime = new Date(Long.MAX_VALUE);
    else
      expirTime = this.expirTime;
    // end if

    if (!startTime.before(to)) return vector.elements();
    if (expirTime.before(from)) return vector.elements();
    Calendar c1 = CalendarCreator.get();
    Calendar c2 = CalendarCreator.get();
    Calendar c3 = CalendarCreator.get();
    int curDist = distance;
    int yearDist;
    int monthDist;
    int dayDist;
    boolean rolledBack;

    switch (periodical) {
      case EVERY_WEEK:
        curDist *= 7;
      case EVERY_DAY:
        c1.setTime(startTime);
        c2.setTime(endTime);
        c3.setTime(from);
        yearDist = c3.get(Calendar.YEAR) - c1.get(Calendar.YEAR);
        if (yearDist > 1) {
          // this operation should increase the performance
          yearDist--;
          dayDist = ((yearDist*365)/curDist)*curDist;
          c1.add(Calendar.DATE,dayDist);
          c2.add(Calendar.DATE,dayDist);
        } // end if
        while (!c2.getTime().after(from)) {
          c1.add(Calendar.DATE,curDist);
          c2.add(Calendar.DATE,curDist);
        } // end while
        while ((c1.getTime().before(to)) &&
               (c1.getTime().before(expirTime))) {
          vector.addElement(makeInstance(c1.getTime(),c2.getTime()));
          c1.add(Calendar.DATE,curDist);
          c2.add(Calendar.DATE,curDist);
        } // end while
      break;
      case EVERY_MONTH:
      case EVERY_MONTH_EXACT: {
        int targetDay;
        Date rolledExpirTime;

        c1.setTime(startTime);
        c2.setTime(endTime);
        targetDay = c1.get(Calendar.DATE);
        if (targetDay > 28) {
          c1.add(Calendar.DATE,-3);
          c2.add(Calendar.DATE,-3);
          c3.setTime(expirTime);
          c3.add(Calendar.DATE,-3);
          rolledExpirTime = new Date(c3.getTime().getTime());
          c3.setTime(to);
          c3.add(Calendar.DATE,-3);
          to = new Date(c3.getTime().getTime());
          c3.setTime(from);
          c3.add(Calendar.DATE,-3);
          from = new Date(c3.getTime().getTime());
          rolledBack = true;
        }
        else {
          rolledExpirTime = expirTime;
          c3.setTime(from);
          rolledBack = false;
        } // end if
        yearDist  = c3.get(Calendar.YEAR) - c1.get(Calendar.YEAR);
        monthDist = 12*yearDist+c3.get(Calendar.MONTH)-c1.get(Calendar.MONTH);
        monthDist = (monthDist/curDist)*curDist;
        if (monthDist > 0) {
          c1.add(Calendar.MONTH,monthDist);
          c2.add(Calendar.MONTH,monthDist);
        } // end if
        while (!c2.getTime().after(from)) {
          c1.add(Calendar.MONTH,curDist);
          c2.add(Calendar.MONTH,curDist);
        } // end while
        while ((c1.getTime().before(to)) &&
               (c1.getTime().before(rolledExpirTime))) {
          if (rolledBack) {
            c1.add(Calendar.DATE,3);
            if (c1.get(Calendar.DATE) == targetDay) {
              c2.add(Calendar.DATE,3);
              vector.addElement(makeInstance(c1.getTime(),c2.getTime()));
              c2.add(Calendar.DATE,-3);
            } // end if
            c1.add(Calendar.DATE,-3);
          }
          else {
            vector.addElement(makeInstance(c1.getTime(),c2.getTime()));
          } // end if
          c1.add(Calendar.MONTH,curDist);
          c2.add(Calendar.MONTH,curDist);
        } // end while
      } break;
      case EVERY_MONTH_MATCH: {
        int targetDay;
        int targetWeek;
        int fromMonth;
        int toMonth;
        int startMonth;
        int i,diff;

        c3.setTime(startTime);
        c3.add(Calendar.DATE,-3);
        i = c3.get(Calendar.DAY_OF_WEEK);
        c1.setFirstDayOfWeek(i);
        c2.setFirstDayOfWeek(i);
        c3.setFirstDayOfWeek(i);

        c1.setTime(startTime);
        c2.setTime(endTime);
        c3.setTime(to);
        toMonth = 12*c3.get(Calendar.YEAR)+c3.get(Calendar.MONTH);
        c3.setTime(from);
        fromMonth  = 12*c3.get(Calendar.YEAR)+c3.get(Calendar.MONTH);
        startMonth = 12*c1.get(Calendar.YEAR)+c1.get(Calendar.MONTH);
        targetDay  = c1.get(Calendar.DAY_OF_WEEK);
        targetWeek = c1.get(Calendar.WEEK_OF_MONTH);
        // set the date to the 15th of month
        diff = 15 - c1.get(Calendar.DATE);
        c1.add(Calendar.DATE,diff);
        c2.add(Calendar.DATE,diff);
        for (i=fromMonth; i<=toMonth; i++) {
          diff = i - startMonth;
          if ((diff < 0) || ((diff % curDist) != 0)) continue;
          if (diff > 0) {
            c1.add(Calendar.MONTH,diff);
            c2.add(Calendar.MONTH,diff);
            startMonth = i;
          } // end if
          // set day of week
          while (c1.get(Calendar.DAY_OF_WEEK) != targetDay) {
            c1.add(Calendar.DATE,-1);
            c2.add(Calendar.DATE,-1);
          } // end while
          // set week of month
          diff = 7 * (targetWeek - c1.get(Calendar.WEEK_OF_MONTH));
          c1.add(Calendar.DATE,diff);
          c2.add(Calendar.DATE,diff);
          if ((c1.get(Calendar.WEEK_OF_MONTH) == targetWeek) && // if week exists
              (c1.getTime().before(to)) &&       // if date matches
              (c2.getTime().after(from)) &&
              ((expirTime.getTime() == INVALID_DATE) ||    // if not expired
               (!expirTime.before(c2.getTime())))) {
            vector.addElement(makeInstance(c1.getTime(),c2.getTime()));
          } // end if
          // undo week set
          c1.add(Calendar.DATE,-diff);
          c2.add(Calendar.DATE,-diff);
          // set the date to the 15th of month
          diff = 15 - c1.get(Calendar.DATE);
          c1.add(Calendar.DATE,diff);
          c2.add(Calendar.DATE,diff);
        } // end for
      } break;
      default:
    } // end switch
    return vector.elements();
  } // end makePeriodicInstances

  public CalendarEntry reconstructParent() {
    if (isPeriodic()) {
      CalendarEntry entry = (CalendarEntry)this.clone();
      entry.startTime     = perStartTime;
      entry.endTime       = perEndTime;
      entry.alertTime     = perAlertTime;
      return entry;
    }
    else {
      return this;
    } // end if
  } // end reconstructParent

  public int getDistance() {

    return distance;
  } // end getDistance

  public void setDistance(int i) {

    distance = i;
  } // end setDistance

  public boolean hasExpirDate() {

    return (expirTime.getTime() != INVALID_DATE);
  } // end hasExpirDate

  public Date getExpirTime() {
    return expirTime;
  } // end getExpirTime

  public void setExpirTime(Date date) {
    expirTime = new Date(date.getTime());
  } // end setExpirTime

  public boolean hasAlertTime() {

    return (alertTime.getTime() != INVALID_DATE);
  } // end hasAlertTime

  public Date getAlertTime() {
    return alertTime;
  } // end getAlertTime

  public void setAlertTime(Date date) {
    alertTime = new Date(date.getTime());
  } // end setAlertTime

  public void setPrivate(boolean privateFl) {
    isPrivateFlag = privateFl;
  } // end setPrivate

  public boolean isPrivate() {

    return isPrivateFlag;
  } // end isPrivate

  public void makePublicView(String user) {
    // this methods hides some fields, if the entry is'nt public and the
    // user is'nt the creator
    if ((!isPrivateFlag) || (user.equals(creator))) return;
    alertTime.setTime(INVALID_DATE);
    description   = "<private>";
    location      = "<private>";
    notes         = "<private>";
  } // end makePublicView

  public CalendarEntry getOldVersion() {

    return oldVersion;
  } // end getOldVersion

  public void setOldVersion(CalendarEntry cEntry) {

    oldVersion = cEntry;
  } // end setOldVersion

  public boolean matches(Date from, Date to) {

    if (isDeleted()) return false;
    if (isPeriodic()) {
      return (from.equals(startTime) || from.after(startTime)) &&
             (from.before(expirTime));
    }
    else {
      return (startTime.equals(from) || startTime.after(from)) &&
             (startTime.before(to));
    } // end if
  } // end matches

  public Object clone() {
  // creates a deep copy of the object
  // the oldVersion and the periodic parent dates are not cloned
    CalendarEntry cEntry = new CalendarEntry();

    cEntry.startTime     = new Date(startTime.getTime());
    cEntry.endTime       = new Date(endTime.getTime());
    cEntry.alertTime     = new Date(alertTime.getTime());
    cEntry.description   = description;
    cEntry.creator       = creator;
    cEntry.persons       = (Vector)persons.clone();
    cEntry.location      = location;
    cEntry.notes         = notes;
    cEntry.periodical    = periodical;
    cEntry.distance      = distance;
    cEntry.expirTime     = new Date(expirTime.getTime());
    cEntry.isPrivateFlag = isPrivateFlag;
    cEntry.oldVersion    = null;
    cEntry.lastInSeq     = lastInSeq;
    cEntry.perStartTime  = null;
    cEntry.perEndTime    = null;
    cEntry.perAlertTime  = null;
    return cEntry;
  } // end clone

  public boolean equals(Object obj) {
  // notes and persons list are not compared because of multiuser
  // update operations
  // oldVersion is also excluded from the compare
    if (!this.getClass().equals(obj.getClass())) return false;
    CalendarEntry calEntry = (CalendarEntry)obj;

    return
      startTime.equals(calEntry.startTime) &&
      endTime.equals(calEntry.endTime) &&
      alertTime.equals(calEntry.alertTime) &&
      description.equals(calEntry.description) &&
      creator.equals(calEntry.creator) &&
      location.equals(calEntry.location) &&
      (periodical == calEntry.periodical) &&
      (distance == calEntry.distance) &&
      expirTime.equals(calEntry.expirTime) &&
      (isPrivateFlag == calEntry.isPrivateFlag);
  } // end equals

} // end CalendarEntry

/* ******************************************************************
   end of file
*********************************************************************/
